/*
** mod_chm -- an apache2 module that provides possibility to explore 
**            file in CHM format.
**
** Copyright (C) 2005 Pavel Frolov <snake4d@pisem.net>
** 
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place, Suite 330, Boston, MA, 02111-1307, USA or visit:
** http://www.gnu.org
*/

#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <ap_config.h>

#include <chm_lib.h>
#include <string.h>

#include "mod_chm.h"
#include "images.h"
#include "html_codes.h"

typedef struct _Buffer{
  unsigned char* buffer;
  unsigned int buf_idx;
  unsigned int buf_size;
  unsigned int buf_len;
  unsigned int offset;
} Buffer;

const int nSyms = sizeof(html_codes)/sizeof(struct HtmlCodes);

int chm_index_element( struct chmFile* f, struct chmUnitInfo* ui, void* data);
char* chm_get_index( struct chmFile* file );
int chm_embeded( request_rec *r );
int chm_parse_index( request_rec *r, struct chmFile* file, char* filename );
char* chm_get_param( const char* tag, const char* param );
char* chm_get_next_tag( struct chmFile* file, struct chmUnitInfo* ui, Buffer* b );
int chm_fill_next_buffer( struct chmFile* file, struct chmUnitInfo* ui, Buffer* b );
char* chm_map_symbols( const char* string );
char chm_map_sym( char* sym);
char* chm_map_ltgt( char* string );

/* The sample content handler */
static int chm_handler(request_rec *r)
{
	apr_file_t* fd;
	apr_size_t nbytes;
	char *fname;
	struct chmFile* file;
	struct chmUnitInfo ui;
	unsigned char buffer[BUF_SIZE];
	unsigned int offset=0, len=0;  
  
    if (strcmp(r->handler, CHM_MAGIC_TYPE)) {
        return DECLINED;
    }
    r->content_type = "text/html";      

	if( !r->header_only ){
		//Route path info
		if( (!r->path_info || !(*r->path_info) || !strcmp( r->path_info, CHM_ROOT)) && !r->args ) {
			fname = strdup( r->filename );
			if( r->path_info || *r->path_info )
				ap_rprintf(r, mainHtml, basename(fname), r->uri, "/");
			else	
				ap_rprintf(r, mainHtml, basename(fname), "","");
			free( fname );
		}else if( !r->args ){
			//Try to find file in archive
			if( (file = chm_open( r->filename )) == NULL)
				return chm_embeded(r);
			// Get file
			if( chm_resolve_object( file, r->path_info, &ui ) != CHM_RESOLVE_SUCCESS){
				chm_close( file );
		    	return chm_embeded(r);
			}
			while( offset < ui.length ){
				if( ui.length - offset < BUF_SIZE ) len =  ui.length - offset;
				else len = BUF_SIZE;

			    len =  (unsigned int)chm_retrieve_object( file, &ui, buffer, offset, len );
		    	offset += len;
				ap_rwrite( buffer, len, r);
			}
			chm_close( file );
			return OK;

		}else if( !strcmp(r->args, CHM_INDEX )){
			if( (file = chm_open( r->filename )) == NULL)
				return HTTP_NOT_FOUND;

			fname = chm_get_index( file );
			if( fname == NULL ){
				chm_close( file );
				return HTTP_NOT_FOUND;
			}
			ap_rputs(tocHeader, r);
			chm_parse_index( r, file, fname );
			chm_close( file );
				
			if(fname != NULL) free( fname );
			ap_rputs(tocFooter,r);
			
		}else if( !strcmp(r->args, CHM_DOWNLOAD )){
			r->content_type = CHM_MAGIC_STREAM;
		
			if( !r->finfo.protection )
			return HTTP_NOT_FOUND;
   	
			if (!r->header_only){
				apr_file_open( &fd, r->filename, APR_READ, 	APR_OS_DEFAULT, r->pool );
				ap_send_fd( fd,	r, 0, r->finfo.size, &nbytes );
				apr_file_close( fd );
			}
		}else if( !strcmp(r->args, CHM_TITLE )){
			fname = strdup( r->filename );
			ap_rprintf(r, titleHtml, basename(fname) );
			free( fname );
		}
	}
	
    return OK;
}

static void chm_register_hooks(apr_pool_t *p)
{
    ap_hook_handler(chm_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA chm_module = {
    STANDARD20_MODULE_STUFF, 
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    NULL,                  /* table of config file commands       */
    chm_register_hooks  /* register hooks                      */
};

int chm_index_element( struct chmFile* f, struct chmUnitInfo* ui, void* data)
{
	char** index = (char**)data;
	
	if( strstr(ui->path, ".hhc") ){
		*index = strdup( ui->path );
		return CHM_ENUMERATOR_SUCCESS;
	}else{
		*index = NULL;
		return CHM_ENUMERATOR_CONTINUE;
	}
}

//find .hhc file
//NULL if not found
char* chm_get_index( struct chmFile* file )
{
	char* index;
	
	if( !chm_enumerate( file, CHM_ENUMERATE_ALL, chm_index_element, &index ) )
					return NULL;
	return index;
}

//Try to get request in embeded resources
int chm_embeded(request_rec *r)
{
	static int size = sizeof(embeded_images)/sizeof(struct EmbededImage);
	int i;
	
	for(i=0; i<size; ++i)
		if( !strcmp(r->path_info, embeded_images[i].name )){
			r->content_type = "image/png";
			ap_rwrite( embeded_images[i].data, embeded_images[i].size, r);
			return OK;
		}

	return HTTP_NOT_FOUND;
}

// 0 - error
int chm_parse_index( request_rec *r, struct chmFile* file, char* filename )
{
  typedef struct _Anchor{
	  char* name;
	  char* link;
  } Anchor;
  struct chmUnitInfo ui;
  Anchor a;
  Buffer bf;
  char *name, *value, *tmpchar;
  char *tag, *next_tag;
  unsigned int i,cur_folder = 0, level=0, first_link=1, ignore_next_ul=0;

	// Get index file
	if( !filename || chm_resolve_object( file, filename, &ui ) != CHM_RESOLVE_SUCCESS){
    	return 0;
  	}

	bf.buffer = (unsigned char*)malloc(BUF_SIZE);
	bf.buf_size = BUF_SIZE;
	bf.buf_len = 0;
	bf.buf_idx = 0;
	bf.offset = 0;
	// Parse index
	next_tag = chm_get_next_tag( file, &ui, &bf );
	while( next_tag != NULL ){
		tag = next_tag;
		next_tag = chm_get_next_tag( file, &ui, &bf );
    	if( !strncasecmp(tag,"UL",2)) {
			if( !ignore_next_ul ){  //Fix </UL><UL>
				if( !cur_folder )
					ap_rprintf(r,"<div style=\"display: block;\" id=\"folder%d\">\n",cur_folder++);
				else 
					ap_rprintf(r,"<div id=\"folder%d\">\n",cur_folder++);
				++level;
			}
	    }else if( !strncasecmp( tag, "/UL", 3) ){
			if( next_tag && !strncasecmp ( next_tag, "UL", 2 )){ //Fix </UL><UL> sequence
				ignore_next_ul = 1;
			}else{
				ignore_next_ul = 0;
				ap_rprintf(r,"</div>\n");
				--level;
			}
    	}else if( !strncasecmp(tag,"LI",2) ){
			ap_rprintf(r,"<p>");
	    }else if( !strncasecmp(tag,"/OBJECT",7) ){
			if( a.name && *(a.name) != 0 ){
				
				for(i=0;i<level-1;++i)
					ap_rprintf(r,"<img src=\"mod_chm_images/blank.png\" width=16 height=22>");

				if( next_tag && !strncasecmp(next_tag,"UL",2) ){
					ap_rprintf(r,"<img src=\"mod_chm_images/plusnode.png\" alt=\"+\" width=16 height=22 onclick=\"toggleFolder('folder%d', this)\" >",cur_folder);
					ap_rprintf(r,"<img src=\"mod_chm_images/folderclosed.png\" alt=\"-\" width=24 height=22 onclick=\"toggleFolder('folder%d', this)\" >",cur_folder);
				}else{
					ap_rprintf(r,"<img src=\"mod_chm_images/blank.png\" width=16 height=22 >");
					ap_rprintf(r,"<img src=\"mod_chm_images/doc.png\" alt=\"*\" width=24 height=22 >");
				}
				if(a.link){
					if( first_link ){
						ap_rprintf(r,"<a id=\"first\" class=\"el\" href=\"%s\" target=\"content\">%s</a></p>\n",a.link,a.name);
						first_link = 0;
					}else 
						ap_rprintf(r,"<a class=\"el\" href=\"%s\" target=\"content\">%s</a></p>\n",a.link,a.name);
				}else 
					ap_rprintf(r,"%s</p>\n",a.name);
			}
			if(a.name != NULL ) free( a.name );
			if(a.link != NULL ) free( a.link );
			a.name = a.link = NULL;
	    }else if( !strncasecmp(tag, "OBJECT", 6) ){
		  a.name = NULL;
		  a.link = NULL;
    	  //Make object, set it as current
    	}else if( !strncasecmp(tag,"param", 5) ){
	      // Set parameter to current object
        	name = chm_get_param( tag, "name");
        	if(!name) name = chm_get_param( tag, "NAME");
	        value = chm_get_param( tag, "value");
	        if(!value) value = chm_get_param( tag, "VALUE");
			if( name ){
	    	    if( !strncasecmp(name,"Name",4) ){ 
					a.name = chm_map_ltgt(chm_map_symbols(value));
				}else if( !strncasecmp(name,"Local",5) ){
					a.link = chm_map_ltgt(chm_map_symbols(value));
				}
			}
			if( name != NULL) free( name );
			if( value != NULL) free( value );
	    }
	   //clear tag
	   free(tag);
 	 }
	 if( bf.buffer != NULL ) free( bf.buffer );

	 return 1;
}

char* chm_get_param( const char* tag, const char* param )
{
  const char e = '=';
  const char c = '"';
  const char *end = tag+strlen(tag);
  char *cur, *beg, *name=NULL;
   
  if( (cur = strstr(tag,param)) != NULL ){
    cur += strlen(param);
    
    while( cur<end && *cur != e ) ++cur; 
    if( cur == end ) return name;
    ++cur;
    while( cur<end && *cur != c ) ++cur; 
    if( cur == end ) return name;
    beg = ++cur;
    while( cur<end && *cur != c ) ++cur;
    if( cur == end ) return name;
	name = (char*)malloc(cur-beg+1);
	strncpy(name,beg, cur-beg );
	name[cur-beg]=0;
  }
  
  return name;
}

char* chm_get_next_tag( struct chmFile* file, struct chmUnitInfo* ui, Buffer* b )
{
  unsigned int tagIdx=0;
  char tag[MAX_TAG+1];
  int quote_flag=0;

  while(1){
    while( b->buf_idx < b->buf_len && b->buffer[b->buf_idx++] != '<' );
 	  if( b->buf_idx == b->buf_len ){
		  if( !(chm_fill_next_buffer(file,ui,b))) return NULL;
	  }else break;
  }

  while(1){
    while( b->buf_idx < b->buf_len && b->buffer[b->buf_idx] != '>' && b->buffer[b->buf_idx] != '"' ) 
	if( tagIdx<MAX_TAG ) tag[tagIdx++] = b->buffer[b->buf_idx++];
	else return NULL;

    if( b->buf_idx == b->buf_len ){  
		 if( !(chm_fill_next_buffer(file,ui,b))) return NULL;
   	}else if( b->buffer[b->buf_idx] == '>' ){
		if( quote_flag ){
			if( tagIdx<MAX_TAG ) tag[tagIdx++] = b->buffer[b->buf_idx++];
			else return NULL;
			continue;
		}else break;
	}else if( b->buffer[b->buf_idx] == '"' ){
		quote_flag = !quote_flag;
		if( tagIdx<MAX_TAG ) tag[tagIdx++] = b->buffer[b->buf_idx++];
		else return NULL;
	}
  }
  	tag[tagIdx] = 0;
	return strdup(tag);	
}

int chm_fill_next_buffer( struct chmFile* file, struct chmUnitInfo* ui, Buffer* b )
{
	if( ui->length - b->offset < b->buf_size ) b->buf_len =  ui->length - b->offset;
	else b->buf_len = b->buf_size;
	b->buf_idx = 0;

	if( !b->buf_len ) return 0;
	b->buf_len =  (unsigned int)chm_retrieve_object( file, ui, b->buffer, b->offset, b->buf_len );
    b->offset += b->buf_len;
	return b->buf_len;
}

char* chm_map_symbols( const char* string )
{
	int cur_pos=0;
	int beg, end, ln;
	int len;
	const char b = '&';
	const char e = ';';
	char *stmp, *sym = NULL, *str=NULL;
	char ch;

	if( string && *string != 0 ){
		str = strdup( string );
		len = strlen( str );
		while( cur_pos<len ){
			while( cur_pos<len && str[cur_pos++] != b);
			if( cur_pos == len ) return str;
			beg = cur_pos;

			while( cur_pos<len && str[cur_pos] != e) ++cur_pos; 
			if( cur_pos == len ) return str;
			++cur_pos;

	  		ln = cur_pos - beg;
			sym = (char*)malloc(ln);
			memcpy( sym, str+beg, ln-1);
			sym[ln-1]=0;
		
			ch = chm_map_sym(sym);

			stmp=str;
			str = (char*)malloc( strlen(str) - strlen(sym) + 2 );
			sprintf(str,"%.*s%c%s", beg-1, stmp, ch, stmp+beg+strlen(sym)+1);
			cur_pos = beg;
			len = strlen(str); 
	  		free(sym);
			free(stmp);
		}
	}
	return str;
}


//Map symbols to char 
//No support for multibyte symbols
char chm_map_sym( char* sym)
{
	int cmp,j,i,f1,f2,t;
	
//Fibonacci search
	f1=f2=1;
	while(f1<nSyms){
		f1 += f2;
		f2 = f1-f2;
	}
	f1 -=f2; f2 -= f1;
	i = nSyms - f1 + 1;

  while( (cmp = i<0 ? 1: strcmp(sym, html_codes[i].name )) != 0 ){
	if( i<0 || cmp > 0 ){
		if( f1 == 1 ) return 32;//Not found
		else{
			i += f2;	f1 -= f2; f2 -= f1;
		}
	}else{
		if( f2 == 0 ) return 32;//Not found
		else{
			i -= f2; t = f1-f2;  f1 = f2; f2=t;
		}
	}
  }

// binary search
/*	f1=0; f2=n-1;
	while(f1<=f2){
		i=(f1+f2)/2;
		if( !(cmp = strcmp( sym, html_codes[i].name )) ) return html_codes[i].code;
		else if( cmp<0 )
			f2 = i-1;
		else 
			f1=i+1;
	}
*/
/*
//	linear search
	for(i = 0; i<n; ++i)
		if( !strcmp( html_codes[i].name, sym ) )
			return (char)html_codes[i].code;
*/
	return (char)html_codes[i].code;
//	return 32;
}


// free old pointer to string
char* chm_map_ltgt( char* string )
{
	char *scur,*str = NULL;
	int len,pos=0;
	
	if( string && *string){
		scur = str = string;
		len = strlen(str);
		//Search for '<' or '>'
		while( *scur ){
			if( *scur == '<' ){
				scur = (char*)malloc( len+4 );
				sprintf(scur,"%.*s&lt;%s", pos, str, str+pos+1);
				len += 3;
				pos += 3;
				free(str);
				str = scur;
				scur = str+pos;
			}else if( *scur == '>' ){
				scur = (char*)malloc( len+4 );
				sprintf(scur,"%.*s&gt;%s", pos, str, str+pos+1);
				len += 3;
				pos += 3;
				free(str);
				str = scur;
				scur = str+pos;				
			}
			++scur;
			++pos;
		}
	}
	return str;
}

